"""
Base manage.main hack which supports being an entry point.

In your project setup.py, add::

    entry_points={
        'console_scripts': [
            'yourproject = yourproject.manage:main',
        ],
    }

Replace your manage.py content with::

    #!/usr/bin/env python
    from crudlfap.manage import main

    if __name__ == '__main__':
        main('yourproject.settings')

Note that it will enable DEBUG env var, which wsgi.py should **not** enable.
"""
import os
import sys
import warnings


def main(settings_module=None):
    os.environ.setdefault(
        'DJANGO_SETTINGS_MODULE',
        settings_module or 'crudlfap_example.settings'
    )

    if 'DEBUG' not in os.environ:
        warnings.warn('DEFAULTING DEBUG=1')
        os.environ.setdefault('DEBUG', '1')

    try:
        from django.core.management import execute_from_command_line
    except ImportError as exc:
        raise ImportError(
            'Couldn\'t import Django. Are you sure it\'s installed and '
            'available on your PYTHONPATH environment variable? Did you '
            'forget to activate a virtual environment?'
        ) from exc
    execute_from_command_line(sys.argv)



"""
Crudlfa+ PasswordView, Become and BecomeUser views.

Crudlfa+ takes views further than Django and are expected to:

- generate their URL definitions and reversions,
- check if a user has permission for an object,
- declare the names of the navigation menus they belong to.
"""

import logging

from crudlfap import shortcuts as crudlfap

from django import http
from django.contrib import auth
from django.contrib import messages
from django.contrib.auth.forms import PasswordChangeForm, SetPasswordForm
from django.utils.translation import gettext_lazy as _


logger = logging.getLogger()


class PasswordView(crudlfap.UpdateView):
    slug = 'password'
    material_icon = 'vpn_key'
    color = 'purple darken-4'
    controller = 'modal'
    action = 'click->modal#open'

    def get_title_submit(self):
        return _('update').capitalize()

    def get_form_class(self):
        if self.object == self.request.user:
            cls = PasswordChangeForm
        else:
            cls = SetPasswordForm
        # This fixes the form messages feature from UpdateView
        return type(cls.__name__, (cls,), dict(instance=self.object))

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs['user'] = self.object
        return kwargs


class BecomeUser(crudlfap.ObjectView):
    urlname = 'su'
    menus = ['object']
    material_icon = 'attach_money'
    color = 'pink darken-4'
    link_attributes = {'data-noprefetch': 'true'}

    def get_title_menu(self):
        return _('become').capitalize()

    def get_object(self, queryset=None):
        user = super().get_object()

        if user:
            user.backend = 'django.contrib.auth.backends.ModelBackend'
        else:
            messages.error(
                self.request,
                'Could not find user {}'.format(self.kwargs['username'])
            )

        return user

    def get(self, request, *a, **k):
        logger.info('BecomeUser by {}'.format(self.request.user))
        become_user_realname = str(self.request.user)
        become_user = request.session.get('become_user', request.user.pk)
        auth.login(request, self.object)
        request.session.setdefault(
            'become_user_realname',
            become_user_realname
        )
        request.session['become_user'] = become_user
        messages.info(
            request,
            _('Switched to user %s') % request.user
        )
        return http.HttpResponseRedirect('/' + self.router.registry.urlpath)


class Become(crudlfap.View):
    urlname = 'su'

    def has_perm(self):
        return 'become_user' in self.request.session

    def get_object(self):
        user = self.model.objects.get(pk=self.request.session['become_user'])
        user.backend = 'django.contrib.auth.backends.ModelBackend'
        return user

    def get(self, request, *a, **k):
        logger.info('Become by {}'.format(self.request.user))
        if 'become_user' not in request.session:
            logger.debug(
                'No become_user in session {}'.format(self.request.user))
            return http.HttpResponseNotFound()

        if 'become_user' in request.session:
            user = self.get_object()
            auth.login(request, user)
            messages.info(
                request,
                _('Switched back to your user %s') % user
            )
            if 'become_user' in request.session.keys():
                del request.session['become_user']
        else:
            messages.warning(
                request,
                _('You are still superuser %s') % self.request.user
            )

        return http.HttpResponseRedirect('/' + self.registry.urlpath)



from .conf import install_optional


def get_installed():
    return ['django.contrib.staticfiles', 'crudlfap', 'myapp'].copy()


def test_install_optional():
    bs3 = {'bootstrap3': {'after': 'crudlfap'}}
    messages = {
        'django.contrib.messages': {
            'before': 'django.contrib.staticfiles'
        }
    }
    nonexistent = {
        'nonexistent': None,
    }
    ctx = {
        'crudlfap_auth': None,
    }
    after = get_installed()
    install_optional([bs3], after)
    expect = ['django.contrib.staticfiles', 'crudlfap', 'bootstrap3', 'myapp']
    assert after == expect, "Failed to insert bootstrap3 after crudlfap"
    before_and_after = get_installed()
    install_optional([bs3, messages], before_and_after)
    expected = ['django.contrib.messages', 'django.contrib.staticfiles',
                'crudlfap', 'bootstrap3', 'myapp']
    assert before_and_after == expected, "Failed before and after"
    nono = get_installed()
    install_optional([nonexistent], nono)
    assert nono == get_installed(), "Non existing app installed"
    mod_attr = get_installed()
    install_optional([ctx], mod_attr)
    expected = ['django.contrib.staticfiles', 'crudlfap', 'myapp',
                'crudlfap_auth']
    assert mod_attr == expected, "Failed to append module attribute reference"



from betterforms.changelist import SearchForm

from django import forms
from django.db import models
from django.utils.translation import gettext as _


class SearchMixin(object):
    def get_search_fields(self):
        if hasattr(self.router, 'search_fields'):
            return self.router.search_fields
        return [
            f.name
            for f in self.model._meta.fields
            if isinstance(f, models.CharField)
        ]

    def get_search_form_class(self):
        if not self.search_fields:
            return

        return type(
            self.model.__name__ + 'SearchForm',
            (SearchForm,),
            dict(
                SEARCH_FIELDS=self.search_fields,
                model=self.model,
                q=forms.CharField(label=_('Search'), required=False)
            )
        )

    def get_search_form(self):
        if self.search_fields:
            self.search_form = self.search_form_class(
                self.request.GET,
                queryset=self.object_list
            )
            self.search_form.full_clean()
        else:
            self.search_form = None
        return self.search_form



"""
The CRUDLFA+ Router is able to generate menus checking perms, generate urls ...

.. note:: Note that you can also use non-database backed models, by inheriting
          from models.Model and setting their Meta.managed attribute to False.
          Then, you can use CRUDLFA+ views and routers.

Menus
-----

A menu is referenced by a short name, and CRUDLFA+ generic views already define
a bunch of them, but you can add your own too:

- ``object``: means the view is for a model instance,
- ``object_detail``: means the view should only be visible from detail view,
- ``model``: means the view applies to a model class, such as list view,
- ``main``: means the view should be in the main menu.

To get the views of a router, for a menu, kwargs such as the object, and with
permissions on request.user use :py:class:`Router.get_menu()`. In Jinja2
templates you can call them with:

.. code-block:: django

    {% set views=view.router.get_menu(
        'object',
        view.request,
        object=view.object
    ) %}

Now that Django can generate a menu after serious the refactoring that brought
us to discover this pattern with Etienne Vidal @ DevNix, we rely on Jinja2 to
refactor the HTML to render those menus.

The menu macro takes a list of views as argument, and also a
unique HTML id it can use to generate the dropdown.

.. code-block:: django

    {% import 'crudlfap.html' as crudlfap %}
    {{ crudlfap.dropdown(views, 'row-actions-' + str(object.pk)) }}
    {# also works, different style: #}
    {{ crudlfap.dropdownbutton(views, 'row-actions-' + str(object.pk)) }}

The above code will generate a Material design dropdown menu with an icon and
the other one as a button with all nice icons, titles, permissions checked, and
so on.  This is used everywhere you see a part of the page that can spawn to a
dropdown. If there is only one matching view, it will display only the button.
"""
from django.apps import apps
from django.conf import settings
from django.urls import path
from django.utils.module_loading import import_string

from .route import Route
from .settings import CRUDLFAP_VIEWS
from .utils import guess_urlfield


crudlfap = apps.get_app_config('crudlfap')  # pylint: disable=invalid-name


class ViewsDescriptor(object):
    def __init__(self, default=None):
        self.default = default or []

    def __get__(self, obj, objtype):
        if getattr(obj, '_views', None) is None:
            if callable(self.default):
                obj.views = self.default()
            else:
                obj.views = self.default
        return obj._views

    def __set__(self, obj, value):
        obj._views = Views(value)


class Views(list):
    def __getitem__(self, key):
        if isinstance(key, int):
            return super().__getitem__(key)

        for view in self:
            if view.urlname == key:
                return view

    def __setitem__(self, key, value):
        if isinstance(key, int):
            return super().__setitem__(key, value)

        for i, view in enumerate(self):
            if view.urlname == key:
                return super().__setitem__(i, value)

    def __delitem__(self, key):
        if isinstance(key, int):
            return super().__delitem__(key)

        for i, view in enumerate(self):
            if view.urlname == key:
                return super().__delitem__(i)


class Router(object):
    """
    Base router for CRUDLFA+ Route.

    .. py:attribute:: model

        Optional model class for this Router and all its views.

    .. py:attribute:: views

        :py:class:`ViewsDescriptor` using
        :py:data:`~crudlfap.settings.CRUDLFAP_VIEWS` by default,
        otherwise your list of views.

        .. note:: The final views list is generated by the
                  :py:meth:`generate_views` method.
    """
    views = ViewsDescriptor(
        getattr(settings, 'CRUDFLAP_VIEWS', CRUDLFAP_VIEWS)
    )

    def __getattr__(self, attr):
        if attr.startswith('get_'):
            raise AttributeError('{} or {}()'.format(attr[4:], attr))

        if hasattr(self, 'get_' + attr):
            return getattr(self, 'get_' + attr)()

        raise AttributeError('{} or get_{}()'.format(attr, attr))

    def get_urlfield(self):
        """
        Return Field name of model for reversing url.

        This will return model ` slug ` field if available or ` pk ` field.

        See ``guess_urlfield()`` for detail.
        """
        return guess_urlfield(self.model)

    def get_namespace(self):
        """Generate namespace for this Router views."""
        if self.model:
            return self.model._meta.model_name

    def get_urlpath(self):
        """Return Model name for urlpath."""
        if self.model:
            return self.model._meta.model_name

    def get_app_name(self):
        """Generate app name for this Router views."""
        return self.model._meta.app_label

    def get_registry(self):
        from crudlfap.site import site
        return site

    def __getitem__(self, urlname):
        """Get a view by urlname."""
        for view in self.views:
            if view.urlname == urlname:
                return view

        raise KeyError(
            'View with urlname {} not in router {}'.format(
                urlname,
                type(self).__name__,
            )
        )

    def get(self, urlname, *args):
        try:
            return self[urlname]
        except KeyError:
            if args:
                return args[0]
            raise

    def generate_views(self, *views):
        """
        Generate views for this router, core of the automation in CRUDLFA+.

        This method considers each view in given args or self.views and returns
        a list of usable views.

        Each arg may be a view class or a dict of attributes with a `_cls` key
        for the actual view class.

        It will copy the view class and bind the router on it in the list this
        returns.

        For example, this would cause two view classes to be returned, if
        self.model is ``Artist``, then ``CreateView`` will be used as parent to
        create ``ArtistCreateView`` and ``DetailView`` will be used to create
        ``ArtistDetailView``, also setting the attribute
        ``extra_stuff='bar'``::

            Router(Artist).generate_views([
                CreateView,
                dict(_cls=DetailView, extra_stuff='bar'),
                ListView.factory(paginate_by=12),
            ])
        """
        kwargs = dict(model=self.model, router=self)
        result = []
        for arg in views or self.views:
            view = arg

            if isinstance(view, str):
                view = import_string(view)

            try:
                if not issubclass(view, Route) and view != Route:
                    view = type(view.__name__, (view, Route), {})
            except Exception as e:
                print('Got an error with view:', view)
                raise e

            view = view.clone(**kwargs)

            result.append(view)
        return result

    def __init__(self, model=None, registry=None, views=None, **attributes):
        """Create a Router for a Model."""

        if model is not None:
            self.model = model

        if registry is not None:
            self.registry = registry

        '''
        if fields is None:
            warnings.warn(
                '{} has no fields, defaulting to __all__ !'.format(self)
            )
            fields = '__all__'

        self.fields = fields
        '''

        for name, value in attributes.items():
            setattr(self, name, value)

        self.views = Views(self.generate_views(*(views or [])))

    def register(self):
        """
        Register to self.registry.

        Also, adds the get_absolute_url() method to the model class if it has
        None, to return the reversed url for this instance to the view of this
        Router with the ``detail`` slug.

        Set get_absolute_url in your model class to disable this feature. Until
        then, you got it for free.

        Also, register this router as default router for its model class in the
        RouterRegistry.
        """
        self.registry[self.model] = self

        for view in self.views:
            if view.urlname == 'detail':
                break
        if not self.views or view.urlname != 'detail':
            return

        if not hasattr(self.model, 'get_absolute_url'):
            def get_absolute_url(self):
                from crudlfap import shortcuts as crudlfap
                return crudlfap.site[type(self)]['detail'].clone(
                    object=self).url
            self.model.get_absolute_url = get_absolute_url

    def get_urlpatterns(self):
        """
        Generate URL patterns for this Router views.
        """
        return [view.urlpattern for view in self.views]

    def get_urlpattern(self):
        return path(self.urlpath, (
            [
                path('/', (
                    [v.urlpattern for v in self.views if v.urlpath],
                    None,
                    None
                ))
            ] + [v.urlpattern for v in self.views if not v.urlpath],
            self.app_name,
            self.namespace,
        ))

    def get_menu(self, name, request, **kwargs):
        """
        Return allowed view objects which have ``name`` in their ``menus``.

        For each view class in self.views which have ``name`` in their
        ``menus`` attribute, instanciate the view class with ``request`` and
        kwargs, call ``has_perm()`` on it.

        Return the list of view instances for which ``has_perm()`` has passed.
        """
        views = []

        for v in self.views:
            if name not in getattr(v, 'menus', []):
                continue

            view = v.clone(request=request, **kwargs)()

            if view.has_perm():
                views.append(view)

        return views

    def get_model(self):
        return None

    def has_perm(self, view):
        """
        View's request.user has_perm call with the view's permission_fullcode.
        """
        return view.request.user.has_perm(view.permission_fullcode, view)

    def get_queryset(self, view):
        """Return the queryset for a view, returns all by default."""
        return self.model.objects.all()

    def get_fields(self, view):
        """Return the list of fields for a user."""
        fields = list(self.model._meta.fields)
        fields += list(self.model._meta.local_many_to_many)
        return [f.name for f in fields]

    def get_allowed_groups(self):
        return []



# flake8: noqa: F401

from .generic import (
    CreateView,
    DeleteView,
    DeleteObjectsView,
    DetailView,
    FormView,
    HistoryView,
    ListView,
    ModelFormView,
    ModelView,
    ObjectFormView,
    ObjectView,
    ObjectsFormView,
    ObjectsView,
    TemplateView,
    UpdateView,
    View,
)



from django.contrib.auth.views import LoginView, LogoutView
from django.utils.translation import gettext_lazy as _

from .registry import Registry
from .route import Route
from .views.generic import TemplateView


site = Registry(
    views=[
        Route.factory(
            LoginView,
            title_html=_('Login'),
            redirect_authenticated_user=True,
            authenticate=False,
        ),
        Route.factory(
            LogoutView,
            authenticate=False,
        ),
        TemplateView.clone(
            template_name='crudlfap/home.html',
            title_heading='',
            urlname='home',
            urlpath='',
            authenticate=False,
        )
    ]
)



from django.contrib.admin.models import LogEntry
from django.contrib.contenttypes.models import ContentType
from django.forms import models as model_forms
from django.utils.translation import gettext as _

from .form import FormMixin
from .model import ModelMixin


class ModelFormMixin(ModelMixin, FormMixin):
    """ModelForm Mixin using readable"""
    menus = ['model']

    def get_form_kwargs(self):
        self.form_kwargs = super().get_form_kwargs()
        if (hasattr(self, 'object')
                and issubclass(self.form_class, model_forms.ModelForm)):

            self.form_kwargs.update({'instance': self.object})
        return self.form_kwargs

    def get_form(self):
        self.form = self.form_class(**self.form_kwargs)

    def get_form_fields(self):
        return self.fields

    def get_form_base(self):
        return None

    def get_form_class(self):
        kwargs = dict()
        if self.form_base:
            kwargs['form'] = self.form_base
        return model_forms.modelform_factory(
            self.model,
            fields=self.form_fields,
            **kwargs
        )

    def get_form_invalid_message(self):
        return '{}: {}: {}'.format(
            _(self.view_label),
            self.model_verbose_name,
            _('failure'),
        ).capitalize()

    def get_form_valid_message(self):
        return '{}: {}'.format(
            _(self.view_label),
            self.form.instance,
        ).capitalize()

    def message_html(self, message):
        """Add the detail url for form.instance, if possible."""
        if getattr(self, 'object', None):
            try:
                url = self.object.get_absolute_url()
            except Exception:
                return message

            return ' '.join((
                message,
                '<a href="{}" class="btn-flat toast-action">{}</a>'.format(
                    url,
                    _('detail').capitalize(),
                ),
            ))
        return message

    def get_log_action_flag(self):
        return False

    def get_log_message(self):
        return _(self.view_label)

    def log_insert(self):
        if not LogEntry:
            return

        if not self.request.user.is_authenticated:
            return

        if not self.log_action_flag:
            return

        if hasattr(self, 'object_list'):
            objects = self.object_list
        else:
            objects = [self.object]

        for obj in objects:
            LogEntry.objects.log_action(
                self.request.user.pk,
                ContentType.objects.get_for_model(self.model).pk,
                obj.pk,
                str(obj),
                self.log_action_flag,
                self.log_message,
            )

    def form_valid(self):
        response = super().form_valid()
        self.log_insert()
        return response



from crudlfap import shortcuts as crudlfap

from .models import Post


class AuthBackend:
    def authenticate(self, *args):
        return None  # prevent auth from this backend

    def has_perm(self, user_obj, perm, obj=None):
        view = obj

        if view.model != Post:
            return False

        user = user_obj
        code = view.permission_shortcode

        if code in ('list', 'detail'):
            return True
        elif code == 'add':
            return user.is_authenticated
        elif code == 'change':
            return view.object.editable(user)
        elif code == 'delete':
            if hasattr(view, 'object'):
                return view.object.editable(user)

            # DeleteObjects relies on get_queryset to secure runtime
            return user.is_authenticated

        return super().has_perm(user_obj, perm, obj)


class PostMixin:
    def get_exclude(self):
        if not self.request.user.is_staff:
            return ['owner']
        return super().get_exclude()


class PostCreateView(PostMixin, crudlfap.CreateView):
    def form_valid(self):
        self.form.instance.owner = self.request.user
        return super().form_valid()


class PostUpdateView(PostMixin, crudlfap.UpdateView):
    pass


class PostListView(crudlfap.ListView):
    def get_filter_fields(self):
        if self.request.user.is_staff:
            return ['owner']
        return []


class PostRouter(crudlfap.Router):
    fields = '__all__'
    icon = 'music'
    model = Post

    views = [
        crudlfap.DeleteObjectsView,
        crudlfap.DeleteView,
        PostUpdateView,
        PostCreateView,
        crudlfap.DetailView,
        PostListView.clone(
            search_fields=['name'],
        ),
    ]

    def get_queryset(self, view):
        qs = self.model.objects.get_queryset()
        if view.request.user.is_superuser:
            return qs
        elif view.permission_shortcode in ('change', 'delete'):
            return qs.editable(view.request.user)
        elif view.permission_shortcode in ('list', 'detail'):
            return qs.readable(view.request.user)
        return qs.none()


PostRouter().register()
